import bz2
import gzip
import pathlib
import zipfile
from urllib.parse import urlparse

import pytest

from scinexus.composable import NotCompleted
from scinexus.io_util import (
    _path_relative_to_zip_parent,
    atomic_write,
    get_format_suffixes,
    is_url,
    iter_line_blocks,
    iter_record_chunks,
    iter_splitlines,
    open_,
    open_url,
    path_exists,
)


@pytest.fixture
def tmp_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("test_io")


@pytest.fixture
def home_file(DATA_DIR, HOME_TMP_DIR):
    """makes a temporary directory with file"""
    fn = "sample.tsv"
    contents = (DATA_DIR / fn).read_text()
    (HOME_TMP_DIR / fn).write_text(contents)
    return str(HOME_TMP_DIR / fn)


@pytest.mark.parametrize("transform", [str, pathlib.Path])
def test_open_home(DATA_DIR, home_file, transform):
    """expands tilde for opening / writing to home"""
    data_path = DATA_DIR / "sample.tsv"
    expect = data_path.read_text()
    with open_(transform(home_file)) as infile:
        got = infile.read()
        assert got == expect


def test_does_not_write_if_exception(tmp_dir):
    """file does not exist if an exception raised before closing"""
    test_filepath = tmp_dir / "Atomic_write_test"
    with pytest.raises(AssertionError), atomic_write(test_filepath, mode="w") as f:
        f.write("abc")
        raise AssertionError
    assert not test_filepath.exists()


@pytest.mark.parametrize("suffix", ["gz", "bz2", "zip", "lzma", "xz"])
def test_writes_compressed_formats(DATA_DIR, tmp_dir, suffix):
    """correctly writes / reads different compression formats"""
    fpath = DATA_DIR / "sample.tsv"
    expect = pathlib.Path(fpath).read_text()
    outpath = tmp_dir / f"{fpath.name}.{suffix}"
    with atomic_write(outpath, mode="wt") as f:
        f.write(expect)

    with open_(outpath) as infile:
        got = infile.read()

    assert got == expect, f"write failed for {suffix}"


def test_atomic_invalid_parent_dir():
    with pytest.raises(OSError), atomic_write("invalid_dir/test.txt") as out:
        out.write("will not work")


def test_rename(tmp_dir):
    """Renames file as expected"""
    test_filepath = tmp_dir / "Atomic_write_test"
    open(test_filepath, "w").close()
    assert test_filepath.exists()
    with atomic_write(test_filepath, mode="w") as f:
        f.write("abc")


def test_atomic_write_noncontext(tmp_dir):
    """atomic write works as more regular file object"""
    path = tmp_dir / "foo.txt"
    zip_path = path.parent / f"{path.name}.zip"
    aw = atomic_write(path, in_zip=zip_path, mode="w")
    aw.write("some data")
    aw.close()
    with open_(zip_path) as ifile:
        got = ifile.read()
    assert got == "some data"


def test_open_handles_bom(tmp_dir):
    """handle files with a byte order mark"""
    text = "some text"

    textfile = tmp_dir / "sample.txt"
    textfile.write_text(text, encoding="utf-8-sig")

    gzip_file = tmp_dir / "sample.txt.gz"
    with gzip.open(gzip_file, "wt", encoding="utf-8-sig") as outfile:
        outfile.write(text)

    bzip_file = tmp_dir / "sample.txt.bz2"
    with bz2.open(bzip_file, "wt", encoding="utf-8-sig") as outfile:
        outfile.write(text)

    zip_file = tmp_dir / "sample.zip"
    with zipfile.ZipFile(zip_file, "w") as outfile:
        outfile.write(textfile, "sample.txt")

    for path in (bzip_file, gzip_file, textfile, zip_file):
        with open_(path) as infile:
            got = infile.read()
            assert got == text, f"failed reading {path}"


@pytest.mark.parametrize("non", [None, ""])
def test_open_empty_raises(non):
    with pytest.raises(ValueError):
        open_(non)


def test_aw_zip_from_path(tmp_dir):
    """supports inferring zip archive name from path"""
    path = tmp_dir / "foo.txt"
    zip_path = path.parent / f"{path.name}.zip"
    aw = atomic_write(zip_path, in_zip=True, mode="w")
    aw.write("some data")
    aw.close()
    with open_(zip_path) as ifile:
        got = ifile.read()
        assert got == "some data"

    path = tmp_dir / "foo2.txt"
    zip_path = path.parent / f"{path.name}.zip"
    aw = atomic_write(path, in_zip=zip_path, mode="w")
    aw.write("some data")
    aw.close()
    with open_(zip_path) as ifile:
        got = ifile.read()
        assert got == "some data"


def test_expanduser(tmp_dir):
    """expands user correctly"""
    home = pathlib.Path("~").expanduser()
    test_filepath = tmp_dir / "Atomic_write_test"
    test_filepath = str(test_filepath).replace(str(home), "~")
    with atomic_write(test_filepath, mode="w") as f:
        f.write("abc")


def test_path_relative_to_zip_parent():
    """correctly generates member paths for a zip archive"""
    zip_path = pathlib.Path("some/path/to/a/data.zip")
    for member in ("data/member.txt", "member.txt", "a/b/c/member.txt"):
        got = _path_relative_to_zip_parent(zip_path, pathlib.Path(member))
        assert got.parts[0] == "data"


@pytest.mark.parametrize(
    ("name", "expect"),
    [
        ("suffixes.GZ", (None, "gz")),
        ("suffixes.ABCD", ("abcd", None)),
        ("suffixes.ABCD.BZ2", ("abcd", "bz2")),
        ("suffixes.abcd.BZ2", ("abcd", "bz2")),
        ("suffixes.ABCD.bz2", ("abcd", "bz2")),
    ],
)
def test_get_format_suffixes_returns_lower_case(name, expect):
    """should always return lower case"""
    assert get_format_suffixes(name) == expect


@pytest.mark.parametrize(
    ("name", "expect"),
    [
        ("no_suffixes", (None, None)),
        ("suffixes.gz", (None, "gz")),
        ("suffixes.abcd", ("abcd", None)),
        ("suffixes.abcd.bz2", ("abcd", "bz2")),
        ("suffixes.zip", (None, "zip")),
    ],
)
def test_get_format_suffixes(name, expect):
    """correctly return suffixes for compressed etc.. formats"""
    assert get_format_suffixes(name) == expect


@pytest.mark.parametrize(
    ("name", "expect"),
    [
        ("no_suffixes", (None, None)),
        ("suffixes.gz", (None, "gz")),
        ("suffixes.abcd", ("abcd", None)),
        ("suffixes.abcd.bz2", ("abcd", "bz2")),
        ("suffixes.zip", (None, "zip")),
    ],
)
def test_get_format_suffixes_pathlib(name, expect):
    """correctly return suffixes for compressed etc.. formats from pathlib"""
    assert get_format_suffixes(pathlib.Path(name)) == expect


@pytest.mark.parametrize(
    ("val", "expect"),
    [
        ({}, False),
        ("not an existing path", False),
        ("(a,b,(c,d))", False),
        ("(a:0.1,b:0.1,(c:0.1,d:0.1):0.1)", False),
        (__file__, True),
        (pathlib.Path(__file__), True),
        (NotCompleted("FAIL", "test", message="none", source="unknown"), False),
    ],
)
def test_path_exists(val, expect):
    """robustly identifies whether an object is a valid path and exists"""
    assert path_exists(val) == expect


def test_open_reads_zip(tmp_dir):
    """correctly reads a zip compressed file"""
    text_path = tmp_dir / "foo.txt"
    with open(text_path, "w") as f:
        f.write("any str")

    zip_path = tmp_dir / "foo.zip"
    with zipfile.ZipFile(zip_path, "w") as zf:
        zf.write(text_path)

    with open_(zip_path) as got:
        assert got.readline() == "any str"


def test_open_writes_zip(tmp_dir):
    """correctly writes a zip compressed file"""
    zip_path = tmp_dir / "foo.txt.zip"

    with open_(zip_path, "w") as f:
        f.write("any str")

    with zipfile.ZipFile(zip_path, "r") as zf:
        name = zf.namelist()[0]
        got = zf.open(name).read()
        assert got == b"any str"


def test_open_zip_multi(tmp_dir):
    """zip with multiple records cannot be opened using open_"""
    text_path1 = tmp_dir / "foo.txt"
    with open(text_path1, "w") as f:
        f.write("any str")

    text_path2 = tmp_dir / "bar.txt"
    with open(text_path2, "w") as f:
        f.write("any str")

    zip_path = tmp_dir / "foo.zip"
    with zipfile.ZipFile(zip_path, "w") as zf:
        zf.write(text_path1)
        zf.write(text_path2)

    with pytest.raises(ValueError):
        open_(zip_path)


def test_open_url_write_exceptions():
    """Test 'w' mode (should raise Exception)"""
    with pytest.raises(Exception):
        open_url("http://example.com/test.txt", mode="w")


def test_open_url_exceptions():
    """non-http(s) address for url (should raise Exception)"""
    with pytest.raises(Exception):
        open_url("ftp://example.com/test.txt")


def test_iter_splitlines_one(tmp_path):
    path = tmp_path / "one-line.txt"
    value = "We have text on one line."
    path.write_text(value)
    got = list(iter_splitlines(path))
    assert got == [value]


@pytest.mark.parametrize("newline", ["\n", "\r\n"])
def test_iter_splitlines_line_diff_newline(tmp_path, newline):
    path = tmp_path / "multi-line.txt"
    value = ["We have some", "text on different lines", "which load"]
    with open_(path, mode="w", newline=newline) as out:
        out.write("\n".join(value))
    got = list(iter_splitlines(path, chunk_size=5))
    assert got == value


@pytest.mark.parametrize("newline", ["\n", "\r\n"])
def test_iter_splitlines_file_endswith_newline(tmp_path, newline):
    path = tmp_path / "multi-line.txt"
    value = ["We have some", "text on different lines", "which load"]
    with open_(path, mode="w", newline=newline) as out:
        out.write("\n".join(value) + "\n")
    got = list(iter_splitlines(path, chunk_size=5))
    assert got == value


def test_iter_splitlines_chunk_size_exceeds_file_size(tmp_path):
    path = tmp_path / "multi-line.txt"
    value = ["We have some", "text on different lines", "which load"]
    path.write_text("\n".join(value))
    got = list(iter_splitlines(path, chunk_size=5_000_000))
    assert got == value


@pytest.mark.parametrize(
    "value",
    [
        "With text\nending on a\nended in newline.",
        "With text\nending\non a\nended in newline.",
    ],
)
def test_iter_splitlines_chunk_endswith_newline(tmp_path, value):
    path = tmp_path / "multi-line.txt"
    value = value.splitlines()
    path.write_text("\n".join(value))
    got = list(iter_splitlines(path, chunk_size=11))
    assert got == value


def test_iter_splitlines_chunk_empty_file(tmp_path):
    path = tmp_path / "zero.txt"
    path.write_text("")
    got = list(iter_splitlines(path))
    assert not got


@pytest.mark.parametrize("transform", [str, pathlib.Path])
def test_iter_splitlines_tilde(home_file, transform):
    expect = pathlib.Path(home_file).expanduser().read_text().splitlines()
    got = list(iter_splitlines(transform(home_file)))
    assert len(got) == len(expect)


def test_iter_line_blocks_correct_size(tmp_path):
    path = tmp_path / "multi-line.txt"
    value = ["We have some", "text on different lines", "which load"]
    path.write_text("\n".join(value))
    got = list(iter_line_blocks(path, num_lines=2, chunk_size=5))
    expect = [value[:2], value[-1:]]
    assert got == expect


def test_iter_line_blocks_empty(tmp_path):
    path = tmp_path / "zero.txt"
    path.write_text("")
    got = list(iter_line_blocks(path, num_lines=2))
    assert not got


def test_iter_line_blocks_one(tmp_path):
    path = tmp_path / "one-line.txt"
    value = "We have text on one line."
    path.write_text(value)
    got = list(iter_line_blocks(path, num_lines=2))
    assert got == [[value]]


def test_iter_line_blocks_none_num_lines(tmp_path):
    path = tmp_path / "multi-line.txt"
    value = ["We have some", "text on different lines", "which load"]
    path.write_text("\n".join(value))
    got = list(iter_line_blocks(path, num_lines=None))
    expect = [value]
    assert got == expect


@pytest.mark.parametrize(
    "url",
    [
        "http://example.com",
        b"file://example.txt",
        pathlib.Path("example.txt").absolute().as_uri(),
    ],
)
def test_is_url(url):
    assert is_url(url)


@pytest.mark.parametrize(
    "url",
    [
        "example.txt",
        pathlib.Path("example.txt"),
        b"example.txt",
        r"D:\foo\example.txt",
    ],
)
def test_not_is_url(url):
    assert not is_url(url)


def test_open_url_local(DATA_DIR, tmp_path):
    """using file:///"""
    file_name = "sample.tsv"
    local_path = DATA_DIR / file_name
    with open_(local_path) as infile:
        local_data = infile.read()

    with open_url(local_path.absolute().as_uri()) as infile:
        remote_data = infile.read()

    assert remote_data.splitlines() == local_data.splitlines()


@pytest.fixture
def gzip_uri(DATA_DIR, tmp_path):
    inpath = DATA_DIR / "sample.tsv"
    data = inpath.read_text()
    outpath = tmp_path / "sample.tsv.gz"
    with open_(outpath, "wb") as outfile:
        outfile.write(data.encode("utf8"))
    return outpath.as_uri()


@pytest.mark.parametrize("mode", ["r", "rb", "rt"])
def test_open_url_gzip_mode(gzip_uri, mode):
    with open_url(gzip_uri, mode=mode) as infile:
        got = infile.read()
    expect_type = bytes if "b" in mode else str
    assert isinstance(got, expect_type)


@pytest.mark.slow
@pytest.mark.parametrize(
    "mode",
    ["r", "rb", "rt", None],
)
@pytest.mark.internet
def test_open_url(DATA_DIR, mode):
    """different open mode's all work"""
    file_name = "formattest.fasta"
    remote_root = "https://github.com/user-attachments/files/20321056/{}.gz"

    with open_(DATA_DIR / file_name, mode=mode) as infile:
        local_data = infile.read()

    with open_url(remote_root.format(file_name), mode=mode) as infile:
        remote_data = infile.read()

    assert remote_data.splitlines() == local_data.splitlines()

    # Test using a ParseResult for url
    with open_url(urlparse(remote_root.format(file_name)), mode=mode) as infile:
        remote_data = infile.read()
    assert remote_data.splitlines() == local_data.splitlines()


@pytest.mark.slow
@pytest.mark.internet
def test_open_url_compressed(DATA_DIR):
    """comparing compressed file handling"""
    file_name = "formattest.fasta.gz"
    remote_root = "https://github.com/user-attachments/files/20321056/{}"

    with open_(DATA_DIR / file_name) as infile:
        local_data = infile.read()

    with open_url(remote_root.format(file_name), mode="rt") as infile:
        remote_data = infile.read()

    assert remote_data.splitlines() == local_data.splitlines()


def test_get_compression_open_no_args():
    from scinexus.io_util import _get_compression_open

    with pytest.raises(ValueError, match="either path or compression"):
        _get_compression_open()


def test_open_via_url(DATA_DIR):
    uri = (DATA_DIR / "sample.tsv").absolute().as_uri()
    with open_(uri) as infile:
        got = infile.read()
    assert len(got) > 0


def test_atomic_write_tmpdir_not_exist(tmp_path):
    from scinexus.io_util import atomic_write

    bad_tmpdir = tmp_path / "nonexistent_tmpdir"
    with pytest.raises(FileNotFoundError, match="does not exist"):
        atomic_write(tmp_path / "test.txt", tmpdir=bad_tmpdir, mode="w")


def test_close_rename_zip_in_zip_none(tmp_path):
    path = tmp_path / "test.txt"
    zip_path = tmp_path / "test.zip"
    aw = atomic_write(path, in_zip=zip_path, mode="w")
    aw._in_zip = None
    with pytest.raises(RuntimeError, match="in_zip path is unexpectedly None"):
        aw._close_rename_zip(aw._tmppath)


def test_atomic_write_exit_without_enter(tmp_path):
    aw = atomic_write(tmp_path / "test.txt", mode="w")
    with pytest.raises(ValueError, match="file object is unexpectedly None"):
        aw.__exit__(None, None, None)


def test_iter_splitlines_url(DATA_DIR):
    uri = (DATA_DIR / "sample.tsv").absolute().as_uri()
    got = list(iter_splitlines(uri))
    assert len(got) > 0


@pytest.mark.parametrize("chunk_size", [1, 16, 64, 1024, 5_000_000])
def test_iter_record_chunks_chunk_size_independence(tmp_path, chunk_size):
    delim = b"\n//"
    data = b"record1\n//record2 is longer\n//record3"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    got = list(iter_record_chunks(path=path, delimiter=delim, chunk_size=chunk_size))
    assert got == [b"record1", b"record2 is longer", b"record3"]


def test_iter_record_chunks_delimiter_spans_chunk_boundary(tmp_path):
    delim = b"\n//"
    data = b"AAAAA" + delim + b"BBBBB" + delim + b"CCCCC"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    chunk_size = data.index(delim) + 1
    got = list(iter_record_chunks(path=path, delimiter=delim, chunk_size=chunk_size))
    assert got == [b"AAAAA", b"BBBBB", b"CCCCC"]


def test_iter_record_chunks_record_larger_than_chunk(tmp_path):
    delim = b">"
    record = b"x" * 1000
    data = delim + record + delim + b"short"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    got = list(iter_record_chunks(path=path, delimiter=delim, chunk_size=16))
    assert got == [b"", record, b"short"]


def test_iter_record_chunks_chunk_size_one(tmp_path):
    delim = b">"
    data = b">a>b>c"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    got = list(iter_record_chunks(path=path, delimiter=delim, chunk_size=1))
    assert got == [b"", b"a", b"b", b"c"]


def test_iter_record_chunks_ends_on_delimiter(tmp_path):
    delim = b"\n//"
    data = b"record1\n//record2\n//"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    got = list(iter_record_chunks(path=path, delimiter=delim, chunk_size=8))
    assert got == [b"record1", b"record2"]


def test_iter_record_chunks_no_delimiter(tmp_path):
    data = b"no delimiter present at all"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    got = list(iter_record_chunks(path=path, delimiter=b">", chunk_size=8))
    assert got == [data]


def test_iter_record_chunks_empty_file(tmp_path):
    path = tmp_path / "empty.bin"
    path.write_bytes(b"")
    got = list(iter_record_chunks(path=path, delimiter=b">"))
    assert got == []


def test_iter_record_chunks_empty_delimiter_raises(tmp_path):
    path = tmp_path / "records.bin"
    path.write_bytes(b"anything")
    with pytest.raises(ValueError, match="delimiter must be non-empty"):
        list(iter_record_chunks(path=path, delimiter=b""))


@pytest.mark.parametrize("compression", ["gz", "bz2"])
def test_iter_record_chunks_compressed(tmp_path, compression):
    data = b">a\nAAA>b\nBBB>c\nCCC"
    path = tmp_path / f"records.bin.{compression}"
    with open_(path, mode="wb") as f:
        f.write(data)
    got = list(iter_record_chunks(path=path, delimiter=b">", chunk_size=4))
    assert got == [b"", b"a\nAAA", b"b\nBBB", b"c\nCCC"]


@pytest.mark.parametrize("chunk_size", [None, 5_000_000])
def test_iter_record_chunks_read_all(tmp_path, chunk_size):
    data = b">a>b>c"
    path = tmp_path / "records.bin"
    path.write_bytes(data)
    got = list(iter_record_chunks(path=path, delimiter=b">", chunk_size=chunk_size))
    assert got == [b"", b"a", b"b", b"c"]


def test_iter_record_chunks_url(tmp_path):
    data = b">a\nAAA>b\nBBB>c\nCCC"
    src = tmp_path / "records.bin"
    src.write_bytes(data)
    uri = src.absolute().as_uri()
    got = list(iter_record_chunks(path=uri, delimiter=b">", chunk_size=4))
    assert got == [b"", b"a\nAAA", b"b\nBBB", b"c\nCCC"]



import bz2
import gzip
import json
import pathlib
import pickle
import shutil

import pytest

import scinexus
from scinexus.composable import NotCompleted
from scinexus.data_store import DataStoreDirectory, ReadOnlyDataStoreZipped
from scinexus.deserialise import deserialise_object
from scinexus.io import (
    DEFAULT_DESERIALISER,
    DEFAULT_SERIALISER,
    compress,
    decompress,
    from_json,
    from_primitive,
    open_data_store,
    pickle_it,
    to_json,
    to_primitive,
)
from scinexus.sqlite_data_store import DataStoreSqlite


@pytest.fixture
def tmp_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("io")


@pytest.fixture(autouse=True)
def workingdir(tmp_dir, monkeypatch):
    monkeypatch.chdir(tmp_dir)


@pytest.fixture
def fasta_dir(DATA_DIR, tmp_dir):
    tmp_dir = pathlib.Path(tmp_dir)
    filenames = DATA_DIR.glob("*.fasta")
    fasta_dir = tmp_dir / "fasta"
    fasta_dir.mkdir(parents=True, exist_ok=True)
    for fn in filenames:
        dest = fasta_dir / fn.name
        dest.write_text(fn.read_text())
    return fasta_dir


@pytest.fixture
def zipped_full(fasta_dir):
    source = fasta_dir
    path = shutil.make_archive(
        base_name=str(source),
        format="zip",
        base_dir=source,
        root_dir=source.parent,
    )
    return ReadOnlyDataStoreZipped(pathlib.Path(path), suffix="fasta")


def test_define_data_store(fasta_dir):
    """returns an iterable data store"""
    found = open_data_store(fasta_dir, suffix=".fasta")
    assert len(found) > 1
    found = open_data_store(fasta_dir, suffix=".fasta", limit=2)
    assert len(found) == 2

    found = list(open_data_store(fasta_dir, suffix=".fasta*"))
    assert len(found) > 2

    with pytest.raises(ValueError):
        open_data_store(fasta_dir, suffix="*")

    with pytest.raises(ValueError):
        open_data_store(fasta_dir)

    with pytest.raises(TypeError):
        open_data_store(fasta_dir, 1)


@pytest.mark.parametrize(
    ("serialiser", "deserialiser"),
    [
        (json.dumps, json.loads),
        (pickle.dumps, pickle.loads),
        (lambda x: x, deserialise_object),
    ],
)
def test_deserialiser(serialiser, deserialiser):
    data = {"1": 1, "abc": [1, 2]}
    deserialised = from_primitive(deserialiser=deserialiser)
    assert deserialised(serialiser(data)) == data


def test_pickle_unpickle_apps():
    data = {"a": [1, 2, 3]}
    pkld = to_primitive() + to_json()
    upkld = from_json() + from_primitive()
    assert upkld(pkld(data)) == data


def test_pickle_it_unpickleable():
    def foo(): ...

    app = pickle_it()
    got = app(foo)
    assert isinstance(got, NotCompleted)


@pytest.mark.parametrize(
    ("comp", "decomp"),
    [(bz2.compress, bz2.decompress), (gzip.compress, gzip.decompress)],
)
def test_compress_decompress(comp, decomp):
    data = pickle.dumps({"1": 1, "abc": [1, 2]})
    decompressor = decompress(decompressor=decomp)
    compressor = compress(compressor=comp)
    assert decompressor(compressor(data)) == data


@pytest.mark.parametrize("data", [{"a": [0, 1]}])
def test_default_serialiser_deserialiser(data):
    s = DEFAULT_SERIALISER(data)
    ds = DEFAULT_DESERIALISER(s)
    assert ds == data


def test_to_json():
    to_j = to_json()
    data = {"a": [0, 1]}
    assert to_j(data) == json.dumps(data)


def test_from_json():
    from_j = from_json()
    assert from_j('{"a": [0, 1]}') == {"a": [0, 1]}


def test_to_from_json():
    to_j = to_json()
    from_j = from_json()
    app = to_j + from_j
    data = {"a": [0, 1]}
    assert app(data) == data
    assert app(data) is not data


def test_open_suffix_dirname(tmp_dir):
    outpath = tmp_dir / "melsubgroup_aln_flydivas_v1.2"
    outpath.mkdir(exist_ok=True)
    dstore = open_data_store(outpath, suffix="txt")
    assert isinstance(dstore, DataStoreDirectory)


def test_open_zipped(zipped_full):
    got = open_data_store(zipped_full.source, mode="r", suffix="fasta")
    assert len(got) == len(zipped_full)
    assert isinstance(got, type(zipped_full))


def test_open_data_store_sqlitedb(tmp_dir):
    path = tmp_dir / "test.sqlitedb"
    dstore = open_data_store(path, mode="w")
    assert isinstance(dstore, DataStoreSqlite)


# Tests for top-level scinexus.open_data_store


def test_toplevel_open_data_store(fasta_dir):
    """top-level open_data_store delegates to scinexus.io"""
    found = scinexus.open_data_store(fasta_dir, suffix="fasta")
    assert isinstance(found, DataStoreDirectory)
    assert len(found) > 1


def test_toplevel_open_data_store_sqlitedb(tmp_dir):
    path = tmp_dir / "test_toplevel.sqlitedb"
    dstore = scinexus.open_data_store(path, mode="w")
    assert isinstance(dstore, DataStoreSqlite)


# Tests for set_summary_display / get_summary_display


def test_get_summary_display_default():
    """default summary display is None"""
    orig = scinexus.get_summary_display()
    try:
        scinexus.set_summary_display(None)
        assert scinexus.get_summary_display() is None
    finally:
        scinexus.set_summary_display(orig)


def test_set_get_summary_display():
    """set then get returns the same function"""
    orig = scinexus.get_summary_display()
    try:
        sentinel = lambda data, *, name: data  # noqa: E731
        scinexus.set_summary_display(sentinel)
        assert scinexus.get_summary_display() is sentinel
    finally:
        scinexus.set_summary_display(orig)


def test_set_summary_display_none():
    """setting None clears the display function"""
    orig = scinexus.get_summary_display()
    try:
        scinexus.set_summary_display(lambda data, *, name: data)
        scinexus.set_summary_display(None)
        assert scinexus.get_summary_display() is None
    finally:
        scinexus.set_summary_display(orig)


def test_summary_display_applied(fasta_dir):
    """a registered display function is called by summary properties"""
    orig = scinexus.get_summary_display()
    calls = []

    def track(data, *, name):
        calls.append(name)
        return data

    try:
        scinexus.set_summary_display(track)
        dstore = scinexus.open_data_store(fasta_dir, suffix="fasta")
        _ = dstore.describe
        assert "describe" in calls
    finally:
        scinexus.set_summary_display(orig)


def test_register_datastore_reader_duplicate_none():
    from scinexus.io import _datastore_reader_map, register_datastore_reader

    assert None in _datastore_reader_map
    with pytest.raises(ValueError, match="already in"):
        register_datastore_reader(None)


def test_register_datastore_reader_non_string():
    from scinexus.io import register_datastore_reader

    with pytest.raises(TypeError, match="is not a string"):
        register_datastore_reader(123)


def test_register_datastore_reader_empty_string():
    from scinexus.io import register_datastore_reader

    with pytest.raises(ValueError, match="white-space"):
        register_datastore_reader("")


def test_register_datastore_reader_duplicate_suffix():
    from scinexus.io import register_datastore_reader

    with pytest.raises(ValueError, match="already in"):
        register_datastore_reader("zip")


def test_open_data_store_unknown_suffix(tmp_dir):
    path = tmp_dir / "test.xyz"
    path.write_text("data")
    with pytest.raises(KeyError):
        open_data_store(path, suffix="fasta")


def test_open_data_store_no_suffix_write_mode(tmp_dir):
    outpath = tmp_dir / "newdir"
    with pytest.raises(ValueError, match="suffix is required"):
        open_data_store(outpath, mode="w")


def test_open_data_store_memory_readonly():
    with pytest.raises(NotImplementedError, match="readonly"):
        open_data_store(":memory:", mode="r")



from __future__ import annotations

import contextlib
import inspect
import json
import re
import reprlib
from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
from functools import singledispatch
from io import TextIOWrapper
from pathlib import Path
from typing import TYPE_CHECKING, overload

from scitrack import get_text_hexdigest  # type: ignore[import-untyped]

from scinexus.deserialise import deserialise_object
from scinexus.io_util import get_format_suffixes, open_
from scinexus.parallel import is_master_process

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable, Iterator
    from typing import Any, Self

    from citeable import CitationBase

NOT_COMPLETED_TABLE = "not_completed"
LOG_TABLE = "logs"
MD5_TABLE = "md5"

# used for log files, not-completed results
_special_suffixes = re.compile(r"\.(log|json)$")

CITATIONS_FILE = "bibliography.citations"

NoneType = type(None)


class Mode(Enum):
    r = "r"
    w = "w"
    a = "a"


APPEND = Mode.a
OVERWRITE = Mode.w
READONLY = Mode.r

# Summary display registry
_summary_display_func: Callable[..., Any] | None = None

# Unique-ID extractor registry
_id_from_source_func: Callable[..., Any] | None = None


def set_summary_display(func: Callable[..., Any] | None) -> None:
    """Set the function used to display data store summaries.

    Parameters
    ----------
    func
        A callable with signature ``func(data, *, name) -> Any`` where
        *data* is a ``dict`` or ``list[dict]`` and *name* identifies the
        summary method (e.g. ``"describe"``). Pass ``None`` to clear.
    """
    global _summary_display_func  # noqa: PLW0603
    _summary_display_func = func


def get_summary_display() -> Callable[..., Any] | None:
    """Return the currently registered summary display function, or ``None``."""
    return _summary_display_func


def _apply_summary_display(data: Any, *, name: str) -> Any:
    if _summary_display_func is not None:
        return _summary_display_func(data, name=name)
    return data


def _summary_property(data_method: Callable[..., Any]) -> property:
    """Create a property that delegates to a protected data method and applies display.

    The *data_method* should be a method defined on ``DataStoreABC`` (or a
    subclass) whose name starts with ``_``.  Subclasses customise the raw
    data by overriding the ``_``-prefixed method; the public property
    created here handles display wrapping automatically.
    """
    method_name = data_method.__name__
    public_name = method_name.removeprefix("_")

    def fget(self: DataStoreABC) -> Any:
        data = getattr(self, method_name)()
        return _apply_summary_display(data, name=public_name)

    return property(fget, doc=data_method.__doc__)


class DataMemberABC(ABC):
    """Abstract base class for DataMember

    A data member is a handle to a record in a DataStore. It has a reference
    to its data store and a unique identifier.
    """

    @property
    @abstractmethod
    def data_store(self) -> DataStoreABC: ...

    @property
    @abstractmethod
    def unique_id(self) -> str: ...

    def __str__(self) -> str:
        return self.unique_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(data_store={self.data_store.source}, unique_id={self.unique_id})"

    def read(self) -> str | bytes:
        return self.data_store.read(self.unique_id)

    def __eq__(self, other: object) -> bool:
        """to check equality of members and check existence of a
        member in a list of members"""
        return isinstance(other, type(self)) and (self.data_store, self.unique_id) == (
            other.data_store,
            other.unique_id,
        )

    @property
    def md5(self) -> str | None:
        return self.data_store.md5(self.unique_id)


class DataStoreABC(ABC):
    """Abstract base class for DataStore"""

    _init_vals: dict[str, Any]
    _completed: list[DataMemberABC]
    _not_completed: list[DataMemberABC]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        obj = object.__new__(cls)

        init_sig = inspect.signature(cls.__init__)
        bargs = init_sig.bind_partial(cls, *args, **kwargs)
        bargs.apply_defaults()
        init_vals = bargs.arguments
        init_vals.pop("self", None)

        obj._init_vals = init_vals
        obj._completed = []
        obj._not_completed = []
        return obj

    @property
    @abstractmethod
    def source(self) -> str | Path:
        """string that references connecting to data store, override in subclass constructor"""
        ...

    @property
    @abstractmethod
    def mode(self) -> Mode:
        """string that references datastore mode, override in subclass constructor"""
        ...

    @property
    @abstractmethod
    def limit(self) -> int | None: ...

    def __repr__(self) -> str:
        name = self.__class__.__name__
        construction = ", ".join(f"{k}={v}" for k, v in self._init_vals.items())
        return f"{name}({construction})"

    def __str__(self) -> str:
        num = len(self.members)
        name = self.__class__.__name__
        sample = f"{list(self[:2])}..." if num > 2 else list(self)
        return f"{num}x member {name}(source='{self.source}', members={sample})"

    @overload
    def __getitem__(self, index: int) -> DataMemberABC: ...
    @overload
    def __getitem__(self, index: slice) -> list[DataMemberABC]: ...
    def __getitem__(self, index: int | slice) -> DataMemberABC | list[DataMemberABC]:
        return self.members[index]

    def __len__(self) -> int:
        return len(self.members)

    def __contains__(self, identifier: object) -> bool:
        """whether relative identifier has been stored"""
        return any(m.unique_id == identifier for m in self)

    @abstractmethod
    def read(self, unique_id: str) -> str | bytes: ...

    def _check_writable(self, unique_id: str) -> None:
        if self.mode is READONLY:
            msg = "datastore is readonly"
            raise OSError(msg)
        if unique_id in self and self.mode is APPEND:
            msg = "cannot overwrite existing record in append mode"
            raise OSError(msg)

    @abstractmethod
    def write(self, *, unique_id: str, data: str | bytes) -> None:
        self._check_writable(unique_id)

    @abstractmethod
    def write_not_completed(self, *, unique_id: str, data: str | bytes) -> None:
        self._check_writable(unique_id)

    @abstractmethod
    def write_log(self, *, unique_id: str, data: str | bytes) -> None:
        self._check_writable(unique_id)

    @property
    def members(self) -> list[DataMemberABC]:
        return self.completed + self.not_completed

    def __iter__(self) -> Iterator[DataMemberABC]:
        yield from self.members

    @property
    @abstractmethod
    def logs(self) -> list[DataMemberABC]: ...

    @property
    @abstractmethod
    def completed(self) -> list[DataMemberABC]: ...

    @property
    @abstractmethod
    def not_completed(self) -> list[DataMemberABC]: ...

    def _summary_logs(self) -> list[dict]:
        """returns a list of dicts summarising log files"""
        rows = []
        for record in self.logs:
            lines = str(record.read()).splitlines()
            first = lines.pop(0).split("\t")
            row = {"time": first[0], "name": record.unique_id}
            key: str | None = None
            mapped: dict[str, str] = {}
            for line in lines:
                parts = line.split("\t")[-1].split(" : ", maxsplit=1)
                if len(parts) == 1:
                    if key is None:
                        msg = "malformed log data: continuation line before any key"
                        raise ValueError(msg)
                    mapped[key] += parts[0]
                    continue

                key = parts[0]
                mapped[key] = parts[1]

            row["python_version"] = mapped["python"]
            row["who"] = mapped["user"]
            row["command"] = mapped["command_string"]
            row["composable"] = mapped.get("composable function", "")
            rows.append(row)
        return rows

    summary_logs = _summary_property(_summary_logs)

    def _summary_not_completed(self) -> list[dict]:
        """returns a list of dicts summarising not completed results"""
        return summary_not_completeds(self.not_completed)

    summary_not_completed = _summary_property(_summary_not_completed)

    def _describe(self) -> dict[str, object]:
        num_not_completed = len(self.not_completed)
        num_completed = len(self.completed)
        num_logs = len(self.logs)
        return {
            "completed": num_completed,
            "not_completed": num_not_completed,
            "logs": num_logs,
        }

    describe = _summary_property(_describe)

    @abstractmethod
    def drop_not_completed(self, *, unique_id: str | None = None) -> None: ...

    def _validate(self) -> dict[str, object]:
        correct_md5 = len(self.members)
        missing_md5 = 0
        for m in self.members:
            data = m.read()
            md5 = self.md5(m.unique_id)
            if md5 is None:
                missing_md5 += 1
                correct_md5 -= 1
            elif md5 != get_text_hexdigest(data):
                correct_md5 -= 1

        incorrect_md5 = len(self.members) - correct_md5 - missing_md5

        return {
            "md5_correct": correct_md5,
            "md5_incorrect": incorrect_md5,
            "md5_missing": missing_md5,
            "has_log": len(self.logs) > 0,
        }

    def validate(self) -> dict[str, object]:
        return _apply_summary_display(self._validate(), name="validate")

    @abstractmethod
    def md5(self, unique_id: str) -> str | None:
        """
        Parameters
        ----------
        unique_id
            name of data store member

        Returns
        -------
        md5 checksum for the member, if available, None otherwise
        """

    def write_citations(self, *, data: tuple[CitationBase, ...]) -> None:
        """Write citations to the data store. Subclasses should override."""
        if not data:
            return
        import warnings

        warnings.warn(
            f"{type(self).__name__!r} does not support saving citations",
            UserWarning,
            stacklevel=2,
        )

    def _summary_citations(self) -> list[dict]:
        """Return a list of dicts summarising stored citations."""
        if type(self)._load_citations is DataStoreABC._load_citations:
            import warnings

            warnings.warn(
                f"{type(self).__name__!r} does not support saving citations",
                UserWarning,
                stacklevel=2,
            )
        citations = self._load_citations()
        return [{"app": c.summary()[0], "citation": c.summary()[1]} for c in citations]

    summary_citations = _summary_property(_summary_citations)

    def write_bib(self, dest_path: str | Path) -> None:
        """Write stored citations as a BibTeX .bib file."""
        citations = self._load_citations()
        if not citations:
            import warnings

            warnings.warn(
                "No citations stored in this data store",
                UserWarning,
                stacklevel=2,
            )
            return
        from citeable import write_bibtex

        dest_path = Path(dest_path).expanduser().absolute()
        write_bibtex(citations, dest_path)

    def _load_citations(self) -> list[CitationBase]:
        """Load stored citations. Override in subclasses."""
        return []


class DataMember(DataMemberABC):
    """Generic DataMember class, bound to a data store. All read operations
    delivered by the parent."""

    def __init__(self, *, data_store: DataStoreABC, unique_id: str) -> None:
        self._data_store = data_store
        self._unique_id = str(unique_id)

    @property
    def data_store(self) -> DataStoreABC:
        return self._data_store

    @property
    def unique_id(self) -> str:
        return self._unique_id


def summary_not_completeds(
    not_completed: list[DataMemberABC],
    deserialise: Callable[..., Any] | None = None,
) -> list[dict]:
    """
    Parameters
    ----------
    not_completed
        list of DataMember instances for notcompleted records
    deserialise
        a callable for converting not completed contents, the result of member.read() must be a json string
    """
    err_pat = re.compile(r"[A-Z][a-z]+[A-Z][a-z]+\:.+")
    types = defaultdict(list)
    indices = "type", "origin"
    num_bytes = 0
    for member in not_completed:
        record = member.read()
        if deserialise:
            record = deserialise(record)
        if isinstance(record, bytes):
            num_bytes += 1
            continue
        record = deserialise_object(record)
        key = tuple(getattr(record, k, None) for k in indices)
        match = err_pat.findall(record.message)
        types[key].append([match[-1] if match else record.message, record.source])

    if num_bytes == len(not_completed):
        return []

    rows = []
    maxtring = reprlib.aRepr.maxstring
    reprlib.aRepr.maxstring = 45
    limit_len = 45
    for key in types:
        msg_list, src_list = list(zip(*types[key], strict=False))
        messages = reprlib.repr(", ".join(m.splitlines()[-1] for m in set(msg_list)))
        sources = ", ".join(s.splitlines()[-1] for s in src_list if s)
        if len(sources) > limit_len:
            idx = sources.rfind(",", None, limit_len) + 1
            idx = idx if idx > 0 else limit_len
            sources = f"{sources[:idx]} ..."
        row = {
            "type": getattr(key[0], "value", key[0]),
            "origin": key[1],
            "message": messages,
            "num": len(types[key]),
            "source": sources,
        }
        rows.append(row)
    reprlib.aRepr.maxstring = maxtring  # restoring original val
    return rows


def _tidy_and_check_suffix(suffix: str | None) -> str:
    """tidies suffix by removing leading wildcards and dots"""
    suffix = suffix or ""
    suffix = re.sub(r"^[\s.*]+", "", suffix)  # tidy the suffix
    if not suffix or suffix == "*":
        msg = "suffix is required for DataStoreDirectory and cannot be just a wildcard"
        raise ValueError(msg)

    return suffix


class DataStoreDirectory(DataStoreABC):
    """data store backed by a directory on the filesystem"""

    def __init__(
        self,
        source: str | Path,
        mode: Mode | str = READONLY,
        suffix: str | None = None,
        limit: int | None = None,
        verbose: bool = False,
    ) -> None:
        self._mode = Mode(mode)
        source = Path(source)
        self._source = source.expanduser()
        self.suffix = _tidy_and_check_suffix(suffix)
        self._verbose = verbose
        self._source_check_create(self._mode)
        self._limit = limit

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, str):
            return False
        if not _special_suffixes.search(item):
            item = f"{item}.{self.suffix}" if self.suffix not in item else item
        return super().__contains__(item)

    def _source_check_create(self, mode: Mode) -> None:
        if not is_master_process():
            return

        sub_dirs = [NOT_COMPLETED_TABLE, LOG_TABLE, MD5_TABLE]
        source = self.source
        if mode is READONLY:
            if not source.exists():
                msg = f"'{source}' does not exist"
                raise OSError(msg)
            return

        if not source.exists():
            source.mkdir(parents=True, exist_ok=True)

        for sub_dir in sub_dirs:
            (source / sub_dir).mkdir(parents=True, exist_ok=True)

    @property
    def source(self) -> Path:
        """path that references the data store"""
        return self._source

    @property
    def mode(self) -> Mode:
        """string that references datastore mode, override in subclass constructor"""
        return self._mode

    @property
    def limit(self) -> int | None:
        return self._limit

    def read(self, unique_id: str) -> str:
        """reads data corresponding to identifier"""
        with open_(self.source / unique_id) as infile:
            return infile.read()

    def drop_not_completed(self, *, unique_id: str | None = None) -> None:
        """remove not-completed records from the directory

        Parameters
        ----------
        unique_id
            if provided, only drop the record with this identifier,
            otherwise drop all not-completed records
        """
        unique_id = (unique_id or "").replace(f".{self.suffix}", "")
        unique_id = f"{unique_id}.json" if unique_id else unique_id
        nc_dir = self.source / NOT_COMPLETED_TABLE
        md5_dir = self.source / MD5_TABLE
        for m in list(self.not_completed):
            if unique_id and not m.unique_id.endswith(unique_id):
                continue

            file = nc_dir / Path(m.unique_id).name
            file.unlink()
            md5_file = md5_dir / f"{file.stem}.txt"
            md5_file.unlink()
            self.not_completed.remove(m)

        if not unique_id:
            Path(self.source / NOT_COMPLETED_TABLE).rmdir()
            # reset _not_completed list to force not_completed function to make it again
            self._not_completed: list[DataMemberABC] = []

    @property
    def logs(self) -> list[DataMemberABC]:
        log_dir = self.source / LOG_TABLE
        return (
            [
                DataMember(data_store=self, unique_id=str(Path(LOG_TABLE) / m.name))
                for m in log_dir.glob("*")
            ]
            if log_dir.exists()
            else []
        )

    @property
    def completed(self) -> list[DataMemberABC]:
        if not self._completed:
            self._completed = []
            suffix = f"*.{self.suffix}"
            for i, m in enumerate(self.source.glob(suffix)):
                if self.limit and i == self.limit:
                    break
                self._completed.append(DataMember(data_store=self, unique_id=m.name))
        return self._completed

    @property
    def not_completed(self) -> list[DataMemberABC]:
        if not self._not_completed:
            self._not_completed = []
            for i, m in enumerate((self.source / NOT_COMPLETED_TABLE).glob("*.json")):
                if self.limit and i == self.limit:
                    break
                self._not_completed.append(
                    DataMember(
                        data_store=self,
                        unique_id=str(Path(NOT_COMPLETED_TABLE) / m.name),
                    ),
                )
        return self._not_completed

    def _write(
        self,
        *,
        subdir: str,
        unique_id: str,
        suffix: str,
        data: str,
    ) -> DataMember | None:
        super().write(unique_id=unique_id, data=data)
        # check suffix compatible with this datastore
        sfx, cmp = get_format_suffixes(unique_id)
        if sfx != suffix:
            unique_id = f"{Path(unique_id).stem}.{suffix}"
            sfx, cmp = get_format_suffixes(unique_id)

        unique_id = (
            unique_id.replace(self.suffix, suffix)
            if self.suffix and self.suffix != suffix
            else unique_id
        )
        if suffix != "log" and unique_id in self:
            return None
        newline = None if cmp else "\n"
        mode = "wt" if cmp else "w"
        with open_(self.source / subdir / unique_id, mode=mode, newline=newline) as out:
            out.write(data)

        if subdir == LOG_TABLE:
            return None
        if subdir == NOT_COMPLETED_TABLE:
            member = DataMember(
                data_store=self,
                unique_id=str(Path(NOT_COMPLETED_TABLE) / unique_id),
            )
        elif not subdir:
            member = DataMember(data_store=self, unique_id=unique_id)

        md5 = get_text_hexdigest(data)
        unique_id = unique_id.replace(suffix, "txt")
        unique_id = unique_id if cmp is None else unique_id.replace(f".{cmp}", "")
        with open_(self.source / MD5_TABLE / unique_id, mode="w") as out:
            out.write(md5)

        return member

    def write(self, *, unique_id: str, data: str) -> DataMember:  # type: ignore[override]
        """writes a completed record ending with .suffix

        Parameters
        ----------
        unique_id
            unique identifier
        data
            text data to be written

        Returns
        -------
        a member for this record

        Notes
        -----
        Drops any not-completed member corresponding to this identifier
        """
        member = self._write(
            subdir="",
            unique_id=unique_id,
            suffix=self.suffix,
            data=data,
        )
        self.drop_not_completed(unique_id=unique_id)
        if member is not None:
            self._completed.append(member)
        return member  # type: ignore[return-value]

    def write_not_completed(self, *, unique_id: str, data: str) -> DataMember:  # type: ignore[override]
        """writes a not completed record as json

        Parameters
        ----------
        unique_id
            unique identifier
        data
            text data to be written

        Returns
        -------
        a member for this record
        """
        (self.source / NOT_COMPLETED_TABLE).mkdir(parents=True, exist_ok=True)
        member = self._write(
            subdir=NOT_COMPLETED_TABLE,
            unique_id=unique_id,
            suffix="json",
            data=data,
        )
        if member is not None:
            self._not_completed.append(member)
        return member  # type: ignore[return-value]

    def write_log(self, *, unique_id: str, data: str) -> None:  # type: ignore[override]
        (self.source / LOG_TABLE).mkdir(parents=True, exist_ok=True)
        _ = self._write(subdir=LOG_TABLE, unique_id=unique_id, suffix="log", data=data)

    def md5(self, unique_id: str) -> str | None:
        """
        Parameters
        ----------
        unique_id
            name of data store member

        Returns
        -------
        md5 checksum for the member, if available, None otherwise
        """
        uid_name = Path(unique_id).name
        md5_name = re.sub(rf"[.]({self.suffix}|json)$", ".txt", uid_name)
        path = self.source / MD5_TABLE / md5_name

        return path.read_text() if path.exists() else None

    def write_citations(self, *, data: tuple[CitationBase, ...]) -> None:
        if not data:
            return
        from citeable import to_jsons

        path = self.source / CITATIONS_FILE
        path.write_text(to_jsons(data))

    def _load_citations(self) -> list[CitationBase]:
        from citeable import from_jsons

        path = self.source / CITATIONS_FILE
        if not path.exists():
            return []
        return from_jsons(path.read_text())


class ReadOnlyDataStoreZipped(DataStoreABC):
    """read-only data store backed by a zip archive"""

    def __init__(
        self,
        source: str | Path,
        mode: Mode | str = READONLY,
        suffix: str | None = None,
        limit: int | None = None,
        verbose: bool = False,
    ) -> None:
        self._mode = Mode(mode)
        if self._mode is not READONLY:
            msg = "this is a read only data store"
            raise ValueError(msg)

        self.suffix = _tidy_and_check_suffix(suffix)
        source = Path(source)
        self._source = source.expanduser()
        if not self._source.exists():
            msg = f"{self._source!s} does not exit"
            raise OSError(msg)

        self._verbose = verbose
        self._limit = limit

    @property
    def limit(self) -> int | None:
        return self._limit

    @property
    def mode(self) -> Mode:
        return self._mode

    @property
    def source(self) -> Path:
        return self._source

    def read(self, unique_id: str) -> str | bytes:
        """reads data corresponding to identifier from the zip archive"""
        import zipfile

        member_path = str(Path(self.source.stem, unique_id)).replace("\\", "/")
        with zipfile.ZipFile(self.source) as archive:
            raw = archive.open(member_path)
            wrapped = TextIOWrapper(raw, encoding="latin-1")
            return wrapped.read()

    def _iter_matches(self, subdir: str, pattern: str) -> Iterator[Path]:
        import zipfile

        with zipfile.ZipFile(self._source) as archive:
            names = archive.namelist()
            for name in names:
                p = Path(name)
                if subdir and p.parent.name != subdir:
                    continue
                if p.match(pattern) and not p.name.startswith("."):
                    yield p

    @property
    def completed(self) -> list[DataMemberABC]:
        if not self._completed:
            pattern = f"*.{self.suffix}"
            self._completed = []
            num_matches = 0
            for name in self._iter_matches("", pattern):
                num_matches += 1
                member = DataMember(data_store=self, unique_id=name.name)
                self._completed.append(member)

                if self.limit and num_matches >= self.limit:
                    break

        return self._completed

    @property
    def not_completed(self) -> list[DataMemberABC]:
        if not self._not_completed:
            self._not_completed = []
            num_matches = 0
            nc_dir_path = Path(NOT_COMPLETED_TABLE)
            for name in self._iter_matches(NOT_COMPLETED_TABLE, "*.json"):
                num_matches += 1
                member = DataMember(
                    data_store=self,
                    unique_id=str(nc_dir_path / name.name),
                )
                self._not_completed.append(member)
                if self.limit and num_matches >= self.limit:
                    break

        return self._not_completed

    @property
    def logs(self) -> list[DataMemberABC]:
        log_dir = Path(LOG_TABLE)
        logs: list[DataMemberABC] = []
        for name in self._iter_matches(LOG_TABLE, "*"):
            m = DataMember(data_store=self, unique_id=str(log_dir / name.name))
            logs.append(m)
        return logs

    def md5(self, unique_id: str) -> str | None:
        """
        Parameters
        ----------
        unique_id
            name of data store member

        Returns
        -------
        md5 checksum for the member, if available, None otherwise
        """
        uid_name = Path(unique_id).name
        md5_name = re.sub(rf"[.]({self.suffix}|json)$", ".txt", uid_name)
        md5_dir = Path(MD5_TABLE)
        for name in self._iter_matches(MD5_TABLE, md5_name):
            m = DataMember(data_store=self, unique_id=str(md5_dir / name.name))
            result = m.read()
            return result if isinstance(result, str) else result.decode()
        return None

    def drop_not_completed(self, *, unique_id: str | None = None) -> None:
        """not supported on read-only zip data stores"""
        msg = "zip data stores are read only"
        raise TypeError(msg)

    def write(self, *, unique_id: str, data: str | bytes) -> None:
        msg = "zip data stores are read only"
        raise TypeError(msg)

    def write_not_completed(self, *, unique_id: str, data: str | bytes) -> None:
        msg = "zip data stores are read only"
        raise TypeError(msg)

    def write_log(self, *, unique_id: str, data: str | bytes) -> None:
        msg = "zip data stores are read only"
        raise TypeError(msg)

    def write_citations(self, *, data: tuple[CitationBase, ...]) -> None:
        msg = "zip data stores are read only"
        raise TypeError(msg)

    def _load_citations(self) -> list[CitationBase]:
        import zipfile

        from citeable import from_jsons

        target = str(Path(self.source.stem, CITATIONS_FILE)).replace("\\", "/")
        try:
            with zipfile.ZipFile(self.source) as archive:
                data = archive.read(target).decode("utf-8")
            return from_jsons(data)
        except KeyError:
            return []


def get_unique_id(name: object) -> str | None:
    """strips any format suffixes from name"""
    if (name := get_data_source(name)) is None:
        return None
    suffixes = ".".join(sfx for sfx in get_format_suffixes(name) if sfx)
    return re.sub(rf"[.]{suffixes}$", "", name)


def set_id_from_source(func: Callable[..., Any] | None) -> None:
    """Register a custom function for extracting unique IDs from data objects.

    The registered function is consulted as the default by
    :meth:`AppBase.as_completed` and :meth:`WriterApp.apply_to` to derive a
    unique identifier for each input, and by :class:`NotCompleted` to
    normalise the ``source=`` keyword on error records. Pass ``None`` to
    clear the registration and restore the built-in :func:`get_unique_id`.

    Parameters
    ----------
    func
        A callable taking a single data object and returning a string
        identifier (or ``None`` if no identifier can be extracted). The
        callable must be picklable if scinexus apps will be executed in
        parallel via ``loky`` / MPI.

    Notes
    -----
    Per-call overrides via the ``id_from_source`` keyword on
    :meth:`as_completed` and :meth:`apply_to` still take precedence over
    the registered function. Register before constructing apps for the
    cleanest behaviour.
    """
    global _id_from_source_func  # noqa: PLW0603
    _id_from_source_func = func


def get_id_from_source() -> Callable[..., Any]:
    """Return the active unique-ID extractor.

    Returns the function previously passed to :func:`set_id_from_source`,
    or :func:`get_unique_id` if nothing has been registered.
    """
    return _id_from_source_func or get_unique_id


@singledispatch
def get_data_source(data: object) -> str | None:
    source = getattr(data, "source", None)
    return None if source is None else get_data_source(source)


@get_data_source.register
def _(data: str) -> str | None:
    return get_data_source(Path(data))


@get_data_source.register
def _(data: Path) -> str | None:
    return data.name


@get_data_source.register
def _(data: dict) -> str | None:
    try:
        source = data.get("info", {})["source"]
    except KeyError:
        source = data.get("source", None)  # noqa
    return get_data_source(source)


@get_data_source.register
def _(data: DataMemberABC) -> str | None:
    return str(data.unique_id)


def make_record_for_json(
    identifier: str, data: Any, completed: bool
) -> dict[str, object]:
    """returns a dict for storage as json"""
    with contextlib.suppress(AttributeError):
        data = data.to_rich_dict()

    data = json.dumps(data)
    return {"identifier": identifier, "data": data, "completed": completed}


def load_record_from_json(data: Any) -> tuple[str, Any, bool]:
    """returns identifier, data, completed status from json string"""
    if isinstance(data, str):
        data = json.loads(data)

    value = data["data"]
    if isinstance(value, str):
        with contextlib.suppress(json.JSONDecodeError):
            value = json.loads(value)
    return data["identifier"], value, data["completed"]



"""scinexus: composable app infrastructure for scientific computing."""

import typing as _typing

from scinexus._version import __version__
from scinexus.composable import (
    AppBase,
    AppType,
    ComposableApp,
    LoaderApp,
    NonComposableApp,
    NotCompleted,
    NotCompletedType,
    WriterApp,
    define_app,
    is_app,
    is_app_composable,
)
from scinexus.progress import (
    Progress,
    ProgressContext,
    get_progress,
    set_progress_backend,
)

if _typing.TYPE_CHECKING:  # pragma: no cover
    from typing import Any

__all__ = [
    "AppBase",
    "AppType",
    "ComposableApp",
    "LoaderApp",
    "NonComposableApp",
    "NotCompleted",
    "NotCompletedType",
    "Progress",
    "ProgressContext",
    "WriterApp",
    "__version__",
    "define_app",
    "get_id_from_source",
    "get_parallel_backend",
    "get_progress",
    "get_summary_display",
    "is_app",
    "is_app_composable",
    "open_",
    "open_data_store",
    "set_id_from_source",
    "set_parallel_backend",
    "set_progress_backend",
    "set_summary_display",
]

_LAZY_IMPORTS: dict[str, tuple[str, str]] = {
    "open_": ("scinexus.io_util", "open_"),
    "open_data_store": ("scinexus.io", "open_data_store"),
    "set_parallel_backend": ("scinexus.parallel", "set_parallel_backend"),
    "get_parallel_backend": ("scinexus.parallel", "get_parallel_backend"),
    "set_summary_display": ("scinexus.data_store", "set_summary_display"),
    "get_summary_display": ("scinexus.data_store", "get_summary_display"),
    "set_id_from_source": ("scinexus.data_store", "set_id_from_source"),
    "get_id_from_source": ("scinexus.data_store", "get_id_from_source"),
}


def __getattr__(name: str) -> "Any":
    if name in _LAZY_IMPORTS:
        module_path, attr = _LAZY_IMPORTS[name]
        import importlib

        mod = importlib.import_module(module_path)
        return getattr(mod, attr)
    msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(msg)
